package inria.paperoles.data;

import java.io.BufferedReader;
import java.io.File;
import java.io.IOException;
import java.io.InputStream;
import java.io.InputStreamReader;
import java.net.InetAddress;
import java.net.ServerSocket;
import java.net.Socket;
import java.util.ArrayList;
import java.util.Date;
import java.util.List;
import java.util.Map;
import java.util.regex.Matcher;
import java.util.regex.Pattern;

import papertoolkit.PaperToolkit;
import papertoolkit.application.config.Constants;
import papertoolkit.events.EventDispatcher;
import papertoolkit.events.EventHandler;
import papertoolkit.events.ExtendedPenEvent;
import papertoolkit.events.PenEvent;
import papertoolkit.events.PenEventType;
import papertoolkit.pattern.PatternPackage;
import papertoolkit.pattern.coordinates.PageAddress;
import papertoolkit.pen.PenSample;
import papertoolkit.units.PatternDots;
import papertoolkit.util.DebugUtils;
import papertoolkit.util.files.FileUtils;


public class DataDispatcher {

	/**
	 * Will listen on this port for text commands.
	 */
	public static final int DEFAULT_PLAINTEXT_PORT = Constants.Ports.BATCH_SERVER;

	/**
	 * Tells the server that a client wishes to exit. Closes the client's handler.
	 */
	public static final String EXIT_COMMAND = "[[exit]]";

	private static final Pattern PATTERN_PEN_ID = Pattern.compile("<penID id=\"(.*?)\".*?>");
	private static final Pattern PATTERN_BEGIN_PAGE = Pattern.compile("<page address=\"(.*?)\".*?>");
	private static final Pattern PATTERN_BEGIN_STROKE = Pattern.compile("<stroke begin=\"(.*?)\".*?>");
	private static final Pattern PATTERN_END_PAGE = Pattern.compile("</page>");
	private static final Pattern PATTERN_END_STROKE = Pattern.compile("</stroke>");
	private static final Pattern PATTERN_SAMPLE = Pattern.compile("<p x=\"(.*?)\" y=\"(.*?)\" f=\"(.*?)\" t=\"(.*?)\".*?/>");

	/**
	 * @param eventEngine
	 */
	public DataDispatcher() {
	}

	
	/**
	 * @param xmlDataFile
	 */
	public void process(File xmlDataFile, EventHandler eventHandler) {
		// parse it like we used to do... in BNet
		DebugUtils.println("BatchEventHandler got the file: " + xmlDataFile);

		// read in the whole request file into a String
		// is this an issue if the xml file is large, say 20MB?
		final StringBuilder requestBuffer = FileUtils.readFileIntoStringBuffer(xmlDataFile);

		
		// extract the pen ID
		final Matcher matcherPenID = PATTERN_PEN_ID.matcher(requestBuffer);
		String penID = "0";
		if (matcherPenID.find()) {
			penID = matcherPenID.group(1);
		}
		
		
		final Matcher matcherPageBegin = PATTERN_BEGIN_PAGE.matcher(requestBuffer);
		final Matcher matcherPageEnd = PATTERN_END_PAGE.matcher(requestBuffer);

		// for simulating realtime
		long lastPenUpTimestamp = 0L;
		final String penName = "Batched Pen";

		EventSorter eventSorter = new EventSorter(eventHandler);
		
		while (matcherPageBegin.find() && matcherPageEnd.find()) {
			// DebugUtils.println("Processing Page: ");

			// location of the opening tag <page ...>
			final int beginTagEndIndex = matcherPageBegin.end();
			final int beginTagStartIndex = matcherPageBegin.start();

			// location of the closing tag </page>
			final int endTagStartIndex = matcherPageEnd.start();
			final int endTagEndIndex = matcherPageEnd.end();

			// extract page address
			final String pageAddress = matcherPageBegin.group(1);
			// DebugUtils.println("Page Address: " + pageAddress);

			// save where we got this ink, so we will know later on...
			final PageAddress address = new PageAddress(pageAddress);

			// extract front and end matter
			// final String beginText = requestBuffer.substring(beginTagStartIndex, beginTagEndIndex);
			// final String endText = requestBuffer.substring(endTagStartIndex, endTagEndIndex);

			// extract text in between the begin and end tags
			final String insideText = requestBuffer.substring(beginTagEndIndex, endTagStartIndex);
			// System.out.println("Internal Text Length: " + insideText.length());
			// System.out.println(insideText);

			final Matcher matcherStrokeBegin = PATTERN_BEGIN_STROKE.matcher(insideText);
			final Matcher matcherStrokeEnd = PATTERN_END_STROKE.matcher(insideText);
			// look through the strokes for this page
			while (matcherStrokeBegin.find() && matcherStrokeEnd.find()) {
				final String strokeTimeStamp = matcherStrokeBegin.group(1);
				final long ts = Long.parseLong(strokeTimeStamp);
				// date/time of the beginning of the stroke!
				DebugUtils.println("New Batched Stroke at Time: " + new Date(ts));

				// samples between the <stroke...></stroke>
				final String strokeSampleText = insideText.substring(matcherStrokeBegin.end(),
						matcherStrokeEnd.start());

				// store all the pen samples
				final List<PenSample> samples = new ArrayList<PenSample>();

				final Matcher matcherSample = PATTERN_SAMPLE.matcher(strokeSampleText);
				while (matcherSample.find()) {
					final String x = matcherSample.group(1);
					final String y = matcherSample.group(2);
					final String f = matcherSample.group(3);
					final String t = matcherSample.group(4);

					// make samples and stuff.... add it to the ink
					// DebugUtils.println(x + " " + y + " f=" + f + " ts=" + t);

					final PenSample sample = new PenSample(Double.parseDouble(x), Double.parseDouble(y),
							Integer.parseInt(f), Long.parseLong(t));
					samples.add(sample);
				}

				// dispatch the whole pen stroke
				// TODO Figure out the handling with multiple pens, etc... at some point
				// Figure out how to handle this in simulated REAL-TIME
				for (int i = 0; i < samples.size(); i++) {
					if (i == 0) {
						long timeDiff = samples.get(i).timestamp - lastPenUpTimestamp;
						if (timeDiff > 1000 ) { // 1 second, then we just wait one second
							timeDiff = 1000L;
						}
						else if (timeDiff < 1 ) { // 1 second, then we just wait one second
							timeDiff = 1L;
						}
	// Comment by Fanis
	// don't get why they had this delay!!!!!!!!!!Probably for the online version?????? 					
/*						try {
							// pause a bit, up to 1 second... before triggering a new down sample
							// this avoids the jitter filtering we have in some of the handler classes...
							// TODO: We should remove that filtering, as we now have it in PenClient =\
							Thread.sleep(timeDiff);
						} catch (InterruptedException e) {
							e.printStackTrace();
						}
*/
						// DebugUtils.println("DOWN");
						eventSorter.handleEvent(		
								new ExtendedPenEvent(penID, pageAddress,penName, samples.get(i),
								PenEventType.DOWN, false));

					} else if (i == samples.size() - 1) {
						// DebugUtils.println("UP");
						eventSorter.handleEvent(new ExtendedPenEvent(penID, pageAddress,penName, samples.get(i),
								PenEventType.UP, false));
						lastPenUpTimestamp = samples.get(i).timestamp;
					} else {
						// DebugUtils.println("P");
						eventSorter.handleEvent(new ExtendedPenEvent(penID,pageAddress, penName, samples.get(i),
								PenEventType.SAMPLE, false));
					}
				}
				
			}
		}
		
		eventSorter.submit();
	}
	
}
