package inria.paperoles.rendering;

import inria.paperoles.gestures.ComplexInkStroke;
import inria.paperoles.gestures.GeneralInkStroke;

import java.awt.Color;
import java.awt.Graphics2D;
import java.awt.Paint;
import java.awt.Shape;
import java.awt.geom.AffineTransform;
import java.awt.geom.Area;
import java.awt.geom.Path2D;
import java.awt.geom.Rectangle2D;
import java.io.File;
import java.util.List;
import java.util.Vector;

import javax.media.jai.TiledImage;

import papertoolkit.pen.ink.Ink;
import papertoolkit.pen.ink.InkStroke;
import papertoolkit.pen.ink.InkUtils;
import papertoolkit.units.Pixels;
import papertoolkit.units.Points;
import papertoolkit.units.Units;
import papertoolkit.units.conversion.PixelsPerInch;
import papertoolkit.util.DebugUtils;
import papertoolkit.util.MathUtils;
import papertoolkit.util.geometry.CatmullRomSpline;
import papertoolkit.util.graphics.GraphicsUtils;
import papertoolkit.util.graphics.ImageUtils;
import papertoolkit.util.graphics.JAIUtils;

/**
 * <p>
 * Renders Ink to a Graphics2D or PDF.
 * </p>
 * <p>
 * <span class="BSDLicense"> This software is distributed under the <a
 * href="http://hci.stanford.edu/research/copyright.txt">BSD License</a>. </span>
 * </p>
 * 
 * @author <a href="http://graphics.stanford.edu/~ronyeh">Ron B Yeh</a> (ronyeh(AT)cs.stanford.edu)
 */
public class StrokeRenderer {

	private Ink ink;

	private boolean invertedColors = false;

	private StrokeRenderingTechnique renderingTechnique = new StrokeRenderingTechnique();
	
	Shape shape, smoothShape;
	double width;

	/**
	 * @param theInk
	 */
	public StrokeRenderer(InkStroke stroke) {		
		realizeRapidRendering(stroke);
	//	realizeSmoothRendering(stroke);
	}
	
	private void realizeRapidRendering(InkStroke stroke){ 
		width = stroke.getWidth();
		
		final Path2D.Double path = new Path2D.Double();
		final double[] x = stroke.getXSamples();
		final double[] y = stroke.getYSamples();
		path.moveTo(x[0], y[0]); 
		for (int i = 1; i < stroke.getNumSamples(); i++) {
			path.lineTo(x[i], y[i]);			
		}
			
		shape = path;
		
	}
	
	private void realizeSmoothRendering(InkStroke stroke){ 
		double maxDistanceBetweenSamples = InkUtils.getMaxDistanceBetweenSamples(stroke);
		width = stroke.getWidth();
			
		boolean largeStroke = maxDistanceBetweenSamples > 10;

		if (largeStroke) {	
			final CatmullRomSpline crspline = new CatmullRomSpline();
			final double[] x = stroke.getXSamples();
			final double[] y = stroke.getYSamples();
			crspline.setPoints(x, y);
			smoothShape = crspline.getShape();
		} 
		else {
			smoothShape = shape;
		}	
	}
	
	public Shape getShape(){
		return shape;
	}
	
	public Shape getSmoothShape(){
		return smoothShape;
	}
	
	public void renderToG2D(Graphics2D g2d) {
		renderToG2D(g2d, null);
	}	
	
	/*
	private List<InkStroke> serializeStrokes(List<InkStroke> strokes){
		Vector<InkStroke> sstrokes = new Vector<InkStroke>();
		
		for (InkStroke stroke : strokes) {
			if(stroke instanceof ComplexInkStroke){ System.err.println("Im here...");
				Vector<GeneralInkStroke> instrokes = ((ComplexInkStroke)stroke).getStrokes(); 
				
				for(InkStroke s: instrokes)	sstrokes.add(s);
			}
			else sstrokes.add(stroke);
		}
		
		return sstrokes;
	}	*/
	
	/**
	 * @param g2d
	 */
	public void renderToG2D(Graphics2D g2d, Paint fillPaint) {
		if (ink == null) {
			DebugUtils.println("Ink Object is NULL");
			return;
		}

		// anti-aliased, high quality rendering
		g2d.setRenderingHints(GraphicsUtils.getBestRenderingHints());

		renderingTechnique.render(g2d, shape, width, fillPaint);
	}


	public boolean inkContainsPoint(double x, double y) {
		
		if(shape.contains(x,y)) return true;
		
		return false;
	}

	public boolean inkContains(Rectangle2D rect) {	
		if(shape.contains(rect)) return true;
		else return false;
	}

}
